module ListBoard (
         module PiecesAndPlayers,
         Board (ListBoard),
         initBoard, find, change, findAll,
         indexBy
       ) where


import Data.Char
import Sort
import PiecesAndPlayers



newtype Board
       = ListBoard [[State]]
       deriving (Eq)


instance Show Board
  where

    showsPrec n (ListBoard sss)
      = let
          writePos n
            = if n > 0 then showsPrec 0 n else id
          showRank n (t : ts)
            = case t of
                Just plp
                  -> writePos n .
                     showsPrec 0 plp .
                     showRank 0 ts
                _
                  -> showRank (n + 1) ts
          showRank n _
            = writePos n
        in
        foldr1 (\ f g -> f . ('/' :) . g) (map (showRank n) sss)


ranks
  :: (a -> Bool) -> [a] -> [[a]]
ranks p
  = let
      op x yss
        | p x
          = [] : yss
        | otherwise
          = case yss of
              (zs : zss)
                -> (x : zs) : zss
              _
                -> [[x]]
    in
    foldr op []

readRank
  :: ReadS [State]
readRank s@(c : cs)
  | onBoard k
    = [
        (foldr1 (.) (replicate k (Nothing :)) rs , t)
      |
        (rs , t) <- readRank cs
      ]
  | otherwise
    = [
        (Just r : rs , t2)
      |
        (r , t1) <- reads s,
        (rs , t2) <- readRank t1
      ]
  where
    k = ord c - 48
readRank _
  = [([] , "")]


instance Read Board
  where

    readsPrec n sss
      = let
          (brd , rest)
            = break (== ' ') sss
          readAll (rank : all)
            = [
                r : rs
              |
                (r , "") <- readRank rank,
                length r == 8,
                rs <- readAll all
              ]
          readAll _
            = [[]]
        in
        [
          (ListBoard rs , rest)
        |
          rs <- readAll (ranks (== '/') brd),
          length rs == 8
        ]


indexBy
  :: (Num a, Enum a, Num b, Enum b)
  => (b -> a -> c) -> [[d]] -> [[(c,d)]]
indexBy f
  = zipWith (\ j -> zipWith (\ i s -> (f i j , s)) [1 ..]) [8, 7 ..]


initBoard
  :: Board
initBoard
  = let
      initFirstRank
        = [Rook, Knight, Bishop, Queen, King, Bishop, Knight, Rook]
      initSecondRank
        = replicate 8 Pawn
      emptyRank
        = replicate 8 Nothing
    in
    ListBoard
      [
        map (Just . Piece Black) initFirstRank
      ,
        map (Just . Piece Black) initSecondRank
      ,
        emptyRank
      ,
        emptyRank
      ,
        emptyRank
      ,
        emptyRank
      ,
        map (Just . Piece White) initSecondRank
      ,
        map (Just . Piece White) initFirstRank
      ]

find
  :: Int -> Int -> Board -> State
find i j (ListBoard brd)
  = brd !! (8 - j) !! (i - 1)

change
  :: State -> Int -> Int -> Board -> Board
change st i j (ListBoard brd)
  = let
      change i j (xs : xss)
        = if j == 8
             then let
                    change i (z : zs)
                      = if i == 1
                           then st : zs
                           else z : change (i - 1) zs
                  in
                  change i xs : xss
             else xs : change i (j + 1) xss
      change i j _
        = []
    in
    ListBoard (change i j brd)

findAll
  :: Player -> Board -> [(Piece , Sq)]
findAll pl (ListBoard brd)
  = let
      inter
        = indexBy square brd
      extract ((sq , s) : ps)
        = (case s of
            Just (Piece pl' p)
              -> if pl == pl'
                    then ((p , sq) :)
                    else id
            _
              -> id) (extract ps)
      extract _
        = []
    in
    sort (inter >>= extract)

